use crate::{instructions, stack::Stack};

fn find_labels(stack: &mut Stack) {
    while stack.program_counter < (stack.program.len() as u16) {
        let label = &stack.program[stack.program_counter as usize];
        stack.program_counter += 1;
        
        if !label.name.starts_with('@') {
            continue;
        }

        if stack.labels[label.data as usize].is_some() {
            eprintln!("Label {} already defined", label.data);
            return;
        }

        if label.data == 0 {
            stack.labels[0] = Some(stack.program_counter as i16);
            continue;
        }

        stack.labels[label.data as usize] = Some(stack.program_counter as i16);
    }

    if stack.labels[0].is_some() {
        stack.program_counter = stack.labels[0].unwrap() as u16;
    } else {
        stack.program_counter = 0;
    }
}

pub fn execute(stack: &mut Stack, mut origin_stack: Option<&mut Stack>, mod_name: &String) {
    find_labels(stack);

    while let Some(instruction) = stack.program.get(stack.program_counter as usize) {
        stack.program_counter += 1;
        
        if instruction.name.starts_with('@') {
            continue;
        }

        match instruction.name.as_str() {
            // Stack operations
            "PUSH" => instructions::push::push(&mut stack.memory, &instruction.data),
            "PICK" => instructions::pick::pick(&mut stack.memory, &instruction.data),
            "POP"  => instructions::pop::pop(&mut stack.memory),
            "DUP"  => instructions::dup::dup(&mut stack.memory),
            "SWAP" => instructions::swap::swap(&mut stack.memory),
            "MEOW" => instructions::meow::meow(&mut stack.memory),
            "DUMP" => instructions::dump::dump(&mut stack.memory),
            "SIZE" => instructions::size::size(&mut stack.memory),

            "#EXPR" => instructions::expr::expr(&mut stack.memory, &instruction.arg),

            // Labels
            "JMP" => instructions::jmp::jmp(&mut stack.labels, &mut stack.program_counter, &instruction.data),
            "JNZ" => instructions::jnz::jnz(&mut stack.memory, &mut stack.labels, &mut stack.program_counter, &instruction.data),
            "JZ"  => instructions::jz::jz(&mut stack.memory, &mut stack.labels, &mut stack.program_counter, &instruction.data),

            // Math operations
            "ADD" => instructions::add::add(&mut stack.memory),
            "SUB" => instructions::sub::sub(&mut stack.memory),
            "MUL" => instructions::mul::mul(&mut stack.memory),
            "DIV" => instructions::div::div(&mut stack.memory),

            // Console operations
            "PUTC" => instructions::putc::putc(&mut stack.memory),
            "GETC" => instructions::getc::getc(&mut stack.memory),
            "SCAN" => instructions::scan::scan(&mut stack.memory),

            // Modules
            "ARGS"  => instructions::args::args(&mut stack.memory, &mut origin_stack, &instruction.data),
            "#EXEC" => instructions::exec::exec(stack, &instruction.arg.clone(), &mod_name),

            // ?
            "NOP"  => continue,
            "EXIT" => instructions::exit::exit(&instruction.data),
            "QUIT" => std::process::exit(0),

            _ => {
                eprintln!("Unknown instruction: {}", instruction.name);
                std::process::exit(2);
            }
        }
    }
}